// RW Bumpmaster
// ==============
// Imageboard Retrieval & Archiving Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/curator/bumpmaster/bumpmaster.hpp
// -Declarations of the baseline RW Bumpmaster sub-library

#pragma once

#include "incls_all.hpp"

//------------------------------------------------------------------------------

using Drop_detail = std::tuple<std::string, fs::path, size_t>;

using Pull_list = std::vector<Drop_detail>;

using Sites_boards_map = std::map<std::string, std::set<Drop_detail>>;

using Sites_boards_vec = std::vector<std::tuple<std::string, Pull_list>>;

//------------------------------------------------------------------------------

namespace rw::curator::bumpmaster {

inline fs::path const drops_fspath{"./dolldrops"};

inline unsigned const msg_width_lim{100};

//------------------------------------------------------------------------------
/** A simple data struct to help with thread's post's information
 */
struct Post_data {  // Lynxchan's JSON tag names:
  string   date;    // creation
  string   msg;     // message
  unsigned id;      // postId
};

//------------------------------------------------------------------------------
/** Assign to a post's data struct from a JSON object
 * @param[in] j
 * @param[out] d
 */
void from_json(Json const& j, Post_data& p);

//------------------------------------------------------------------------------
/** Assign to a JSON object from a post's data struct
 * @param[out] j
 * @param[in] d
 */
void to_json(Json& j, Post_data const& p);

//------------------------------------------------------------------------------
/** A simple data struct to help with catalog's thread's information
 */
struct Thrd_data {  // Lynxchan's JSON tag names:
  string   msg;     // message
  string   subj;    // subject
  unsigned id;      // threadId

  vector<Post_data> posts;
};

//------------------------------------------------------------------------------
/** Assign to a thread's data struct from a JSON object
 * @param[in] j
 * @param[out] t
 */
void from_json(Json const& j, Thrd_data& t);

//------------------------------------------------------------------------------
/** Assign to a JSON object from a thread's data struct
 * @param[out] j
 * @param[in] t
 */
void to_json(Json& j, Thrd_data const& t);

//------------------------------------------------------------------------------
/**
 */
class Bumpmaster_b : public RW {
 public:
  Bumpmaster_b();                                    // default ctor
  explicit Bumpmaster_b(std::string const& handle);  // tag-parm'd ctor

  /** Prepare a set of IB archival filesystem directories. Load a container w/
   * these site/board names, and their matching filesystem path's info
   *
   * @param[in] board_uris the source listing of IB URIs to use
   * @return the new info container
   * @note: the results container is sorted by sites (first list
   * appearance), then by boards (first list appearance)
   */
  auto preflight_boards(std::vector<std::string> const& board_uris)
      -> Sites_boards_vec;

  /** Perform some basic formatting operations on a JSON message field
   * @param[in] msg the source to format
   * @return the formatted message
   */
  auto fmt_msg(std::string const& msg) -> std::string;

 protected:
  /** Assemble an IB catalog JSON URI from key parts
   *
   * @param[in] host Internet host name, eg: "alogs.space"
   * @param[in] path Internet path name, eg: "robowaifu"
   * @return the assembled URI
   * @note: example resulting form: https://alogs.space/robowaifu/catalog.json
   */
  auto bld_cat_uri(std::string const& host, std::string const& path)
      -> std::string;

  /** Assemble an IB thread JSON URI from key parts
   *
   * @param[in] host Internet host name, eg: "alogs.space"
   * @param[in] path Internet path name, eg: "robowaifu"
   * @param[in] thrd_id the thread's ID (post number), prev. pulled from catalog
   * @return the assembled URI
   * @note: example resulting form: https://alogs.space/robowaifu/res/14409.json
   */
  auto bld_thrd_uri(std::string const& host,
                    std::string const& path,
                    unsigned const     thrd_id) -> std::string;

  /** Load a container with an IB catalog's thread's data.
   *
   * The local catalog JSON is parsed, and it's thread's data (msg, subj, ID)
   * are stored out.
   *
   * @param[in] cat_json the source JSON data of the IB catalog
   * @return the new container of parsed thread data
   */
  auto bump_thrd_dats(Json const& cat_json) -> std::vector<Thrd_data>;

  /** Load a container with an IB board's thread's JSON URIs
   *
   * @param[in] cat_json the source JSON data of the IB catalog
   * @param[in] site the IB site name to use
   * @param[in] board the IB board name to use
   * @return the new container of URI strings
   */
  auto bump_thrd_uris(Json const&        cat_json,
                      std::string const& site,
                      std::string const& board) -> std::vector<std::string>;

  /** Load a container with an IB thread's post data.
   *
   * The local thread JSON is read and it's post data is extracted, in
   * bump-order.
   *
   * @param[in] thrd_json the source JSON data of the IB thread
   * @return the new container of thread posts data
   */
  auto thrd_posts(Json const& thrd_json) -> std::vector<Post_data>;

  /** Consolidate the original site orderings from the @a board_uris listing
   * @param[in] board_uris the source listing of IB URIs
   * @return listing of sites, once per; in order of appearance
   */
  auto site_order(std::vector<std::string> const& board_uris)
      -> std::vector<std::string>;

  /** Load a map container result with all the sites + their respective boards
   * @param[in] board_uris the source listing of IB URIs
   * @return the new container
   */
  auto map_sites_boards(std::vector<std::string> const& board_uris)
      -> Sites_boards_map;

  /** Sort boards by their pre-stored, original 'board_uris' listing index
   * @param[in/out] io_sites_vec the source container to be sorted
   */
  void sort_boards(Sites_boards_vec& io_sites_vec);

  /** Copy out a @c std::map of IB sites+board URIs into @c std::vector, sorting
   * it by the original 'board_uris' list ordering, into the results container
   *
   * @param[in] sites_map the original source container
   * @param[in] board_uris the source listing of IB URIs to use for sorting
   * @return the new sorted container
   */
  auto sites_map_2_sort_vec(Sites_boards_map const&         sites_map,
                            std::vector<std::string> const& board_uris)
      -> Sites_boards_vec;

  /** Iterate the data container @a io_sites_boards, and ensure the correct
   * directories are created beforehand for the site+board-specific dolldrops
   *
   * @param[in/out] io_sites_boards the source container w/ sites/boards names
   * @return true if any directories needed creating
   * @note: updates each board's directory fspath in the @c Sites_boards_vec
   * container
   */
  auto chk_dirs(Sites_boards_vec& io_sites_boards) -> bool;

  /** Iterate the data container @a io_sites_boards, and overwrite IB board's
   * name field w/ it's newly-built catalog JSON URI string
   * @param[in/out] io_sites_boards the source container getting updated
   * @note: just an expedient utility hack for reusing the container further on
   */
  void brd_2_cat_uri(Sites_boards_vec& io_sites_boards);
};

//------------------------------------------------------------------------------
/**
 */
class Bumpmaster : public Bumpmaster_b {
 public:
  Bumpmaster();                                    // default ctor
  explicit Bumpmaster(std::string const& handle);  // tag-parm'd ctor

  ///

  inline auto waifusearch() const -> Waifusearch& { return waifusearch_; }

 private:
  unique_ptr<Waifusearch_b> _waifusearch_{
      make_unique<Waifusearch_b>("bumpmaster_waifusearch")};

  Waifusearch& waifusearch_{reference_to<Waifusearch>(_waifusearch_.get())};

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

 public:
  /** Switchable signal to the @c Dolldrop member this is a catalog pull request
   *
   * @param[in] flag; default==true
   * @note: (indirectly) affects the @c Dolldrop_b::chkd_pull_g() algorithm
   * @note: simply a wrapper for same function in @c Dolldrop member
   */
  void do_catalog(bool const flag = true);

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (to location specified in @a droppath)
   *
   * @param[in] uri the URI to attempt a download connection with
   * @param[in] droppath the disk file location to save transfer body to
   * @return parsed JSON from @a droppath disk file
   * @note: simply a wrapper for same function in @c Dolldrop member
   */
  auto pull_to_path(std::string const& uri, fs::path const& droppath) -> Json;

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (generic 'anons_file' form)
   *
   * @param[in] uri the URI to attempt a download connection with
   * @return header stream data for @a uri, if present
   * @note: simply a wrapper for same function in @c Dolldrop member
   */
  auto pull_hdr_body(std::string const& uri) -> std::stringstream;

  /** Load a container with an IB board's Page 1 thread's posts; in bump-order.
   *
   * The local catalog JSON is read and it's thread IDs extracted. The top (@a
   * thrd_lim) threads are then check-pulled as dolldrops. These local IB thread
   * JSON files are also parsed out, and their post message's text is then
   * loaded into the results container's matching thread elements.
   *
   * @param[in] cat_json the source JSON data of the IB catalog
   * @param[in] cat_uri the IB catalog's JSON URI (used to gen thread URI's)
   * @param[in] board_jsons_path the local filesystem path to store thread JSONs
   * @param[in] thrd_lim the limit count of threads to parse for this board
   * @return the consolidated threads + posts info
   * @note: this marshaling function effectively merges/unifies thread's data
   * from catalog JSON, with post's data from thread JSON
   */
  auto top_bump_thrds(Json const&        cat_json,
                      std::string const& cat_uri,
                      fs::path const&    board_jsons_path,
                      const int          thrd_lim) -> std::vector<Thrd_data>;

  ///

  // inline auto dolldrop() const -> Dolldrop& { return dolldrop_; }

 private:
  unique_ptr<Dolldrop_b> _dolldrop_{
      make_unique<Dolldrop_b>(t(), "bumpmaster_dolldrop")};

  Dolldrop& dolldrop_{reference_to<Dolldrop>(_dolldrop_.get())};

#endif  // __has_include(<curl_easy.h>)
};

}  // namespace rw::curator::bumpmaster

//    --=^=--

// "______"

// rw_curator_lib.v220507
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
