// RW Bumpmaster
// ==============
// Imageboard Retrieval & Archiving Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/curator/bumpmaster/bumpmaster.cpp
// -Definitions of the baseline RW Bumpmaster sub-library

#pragma once

#include "incls_all.hpp"

namespace rw::curator::bumpmaster {

//------------------------------------------------------------------------------
void from_json(Json const& j, Thrd_data& t) {  //

  if (! j["message"].is_null())
    t.msg = j["message"];
  else
    t.msg = "";

  if (! j["subject"].is_null())
    t.subj = j["subject"];
  else
    t.subj = "";

  if (! j["threadId"].is_null())
    t.id = j["threadId"];
  else
    t.id = 0;
}

//------------------------------------------------------------------------------
void to_json(Json& j, Thrd_data const& t) {
  j = Json{{"message", t.msg}, {"subject", t.subj}, {"threadId", t.id}};
}

//------------------------------------------------------------------------------
void from_json(Json const& j, Post_data& p) {  //

  if (! j["creation"].is_null())
    p.date = j["creation"];
  else
    p.date = "";

  if (! j["message"].is_null())
    p.msg = j["message"];
  else
    p.msg = "";

  if (! j["postId"].is_null())
    p.id = j["postId"];
  else
    p.id = 0;
}

//------------------------------------------------------------------------------
void to_json(Json& j, Post_data const& p) {
  j = Json{{"creation", p.date}, {"message", p.msg}, {"postId", p.id}};
}

//------------------------------------------------------------------------------
Bumpmaster_b::Bumpmaster_b() : Bumpmaster_b{"bumpmaster_b_obj"} {}

//------------------------------------------------------------------------------
Bumpmaster_b::Bumpmaster_b(std::string const& handle)
    : RW{handle, Type::Bumpmaster} {
  stamp("Bumpmaster_b()");
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bld_cat_uri(std::string const& host, std::string const& path)
    -> std::string {
  stamp("bld_cat_uri()");

  string const scheme{"https://"};
  string const page_nm{"catalog"};
  string const type{"json"};

  return {scheme + host + '/' + path + '/' + page_nm + '.' + type};
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bld_thrd_uri(std::string const& host,
                                std::string const& path,
                                unsigned const     thrd_id) -> std::string {
  stamp("bld_thrd_uri()");

  string const scheme{"https://"};
  string const res{"res"};
  string const type{"json"};

  // DESIGN: for now we'll just adapt to Lynxchan's formats for various things
  // (like thread URIs), since our own site's servers use it. later, we'll
  // expand out to every mainline IB type using an approach mostly similar to
  // the one previously devised for BUMP software.

  return {scheme + host + '/' + path + '/' + res + '/' + to_string(thrd_id) +
          '.' + type};
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bump_thrd_dats(Json const& cat_json)
    -> std::vector<Thrd_data> {
  stamp("bump_thrd_dats()");

  vector<Thrd_data> res;
  res.reserve(cat_json.size());

  for (auto const& [idx, cat_thrd] : cat_json.items()) {
    res.emplace_back(cat_thrd);  // from_json() ctors Thrd_data struct, in-place
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bump_thrd_uris(Json const&        cat_json,
                                  std::string const& site,
                                  std::string const& board)
    -> std::vector<std::string> {
  stamp("bump_thrd_uris()");

  vector<string> res;
  res.reserve(cat_json.size());

  // cycle the catalog's threads
  for (auto const& [idx, cat_thrd] : cat_json.items()) {  //

    // build the thread's JSON URI
    auto const thrd_id  = cat_thrd["threadId"];
    auto const thrd_uri = bld_thrd_uri(site, board, thrd_id);

    res.push_back(thrd_uri);
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::thrd_posts(Json const& thrd_json) -> std::vector<Post_data> {
  stamp("thrd_posts()");

  vector<Post_data> res;

  auto posts = thrd_json["posts"];
  res.reserve(posts.size());

  for (auto const& [post_idx, post_dat] : posts.items()) {
    try {
      res.emplace_back(post_dat);  // will ctor a Post_data struct in-place

    } catch (Json::exception const& e) {
      puts("\n\n  caught Json exception in thrd_posts()\n\n");
    }
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::fmt_msg(std::string const& msg) -> std::string {
  stringstream iss{msg}, oss{};

  for (string line; getline(iss, line);) {  // read lines in one at a time
                                            //
    if (line.length() > msg_width_lim) {    // this line needs linebreaks
      // transfer the long line into a sstream
      stringstream liness{line};
      line.clear();

      // parse sstream into multi-lines
      //
      string word{}, curr_line{};
      while (liness >> word) {  // one whitespace-delim'd word at a time

        if (curr_line.length() >= msg_width_lim) {  // we have a linefull, save
          line += (curr_line + '\n');
          curr_line.clear();
        }

        curr_line += (word + ' ');
      }

      trim_r(curr_line, ' ');  // rm final space
      line += curr_line;       // final line pickup for the multi-line
    }

    oss << (line + '\n');
  }

  // rm final newline
  auto res = oss.str();
  trim_r(res, '\n');

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::site_order(std::vector<std::string> const& board_uris)
    -> std::vector<std::string> {  //

  vector<string> sites{};

  for (auto const& board_uri : board_uris) {
    auto const& [site_nm, board_nm] = parse_ib_uri(board_uri);

    if (auto it = find(begin(sites), end(sites), site_nm);
        it == end(sites))  // not found, add it (once; in order of appearance)
      sites.push_back(site_nm);
  }

  return sites;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::map_sites_boards(std::vector<std::string> const& board_uris)
    -> Sites_boards_map {  //

  Sites_boards_map sites_boards;

  size_t orig_idx{0};
  for (auto const& board_uri : board_uris) {
    auto const& [site_nm, board_nm] = parse_ib_uri(board_uri);

    sites_boards[site_nm].insert(make_tuple(board_nm, "", orig_idx++));
  }

  return sites_boards;
}

//------------------------------------------------------------------------------
void Bumpmaster_b::sort_boards(Sites_boards_vec& io_sites_vec) {  //

  for (auto& [site_nm, boards] : io_sites_vec) {  //

    sort(begin(boards), end(boards), [](auto const& A, auto const& B) {
      auto const& [uri_a, path_a, orig_idx_a]{A};
      auto const& [uri_b, path_b, orig_idx_b]{B};

      return orig_idx_a < orig_idx_b;
    });
  }
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::sites_map_2_sort_vec(
    Sites_boards_map const&         sites_map,
    std::vector<std::string> const& board_uris) -> Sites_boards_vec {  //

  Sites_boards_vec res;
  res.reserve(sites_map.size());

  // sort the results container by the original 'board_uris' list ordering
  //
  // -first, sort the sites themselves:
  auto ordered_sites = site_order(board_uris);
  for (auto const& ord_site_nm : ordered_sites) {      // original appearance
                                                       //
    for (auto const& [site_nm, boards] : sites_map) {  // map's sort ordering

      if (site_nm == ord_site_nm) {  // curr elem matches original appearance
        res.emplace_back(
            make_tuple(site_nm, Pull_list{cbegin(boards), cend(boards)}));

        break;  // name already found, so leave inner map loop
      }
    }
  }
  // -next, sort the boards (within each site), as well:
  sort_boards(res);

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::chk_dirs(Sites_boards_vec& io_sites_boards) -> bool {
  bool res{false};

  fs::create_directory(drops_fspath);  // ensure dolldrops directory is ready

  for (auto const& [site_nm, boards] : io_sites_boards) {  // cycle sites
    auto const site_fspath{drops_fspath / site_nm};
    res |= fs::create_directory(site_fspath);

    for (auto const& [board_nm, board_path, idx] : boards) {  // cycle boards
      auto const board_fspath{site_fspath / board_nm};
      res |= fs::create_directory(board_fspath);

      // write the board's inst'd fspath out to it's data item in the container
      const_cast<fs::path&>(board_path) = board_fspath;

      auto const jsons_fspath{board_fspath / "all_jsons"};
      res |= fs::create_directory(jsons_fspath);

      // DESIGN: further dirs (media/thumbs/.static/etc) can be added here
      // ...
    }
  }

  return res;

  // https://en.cppreference.com/w/cpp/language/const_cast
}

//------------------------------------------------------------------------------
void Bumpmaster_b::brd_2_cat_uri(Sites_boards_vec& io_sites_boards) {  //

  for (auto const& [site_nm, boards] : io_sites_boards) {   //
                                                            //
    for (auto const& [board_nm, cat_path, idx] : boards) {  //
      auto cat_uri = bld_cat_uri(site_nm, board_nm);

      const_cast<std::string&>(board_nm) = cat_uri;  // overwrite w/ catalog URI
    }
  }
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::preflight_boards(std::vector<std::string> const& board_uris)
    -> Sites_boards_vec {  //

  auto const sites_map      = map_sites_boards(board_uris);
  auto       sites_sort_vec = sites_map_2_sort_vec(sites_map, board_uris);

  chk_dirs(sites_sort_vec);       // first, check directories using board name
                                  //
  brd_2_cat_uri(sites_sort_vec);  // next, overwrite board name w/ catalog URI

  return sites_sort_vec;
}

//------------------------------------------------------------------------------
Bumpmaster::Bumpmaster() : Bumpmaster{"bumpmaster_obj"} {}

//------------------------------------------------------------------------------
Bumpmaster::Bumpmaster(std::string const& handle) : Bumpmaster_b{handle} {
  stamp("Bumpmaster()");
}

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

//------------------------------------------------------------------------------
void Bumpmaster::do_catalog(bool const flag) {
  stamp("do_catalog()");

  dolldrop_.do_catalog(flag);
}

//------------------------------------------------------------------------------
auto Bumpmaster::pull_to_path(std::string const& uri, fs::path const& droppath)
    -> Json {
  return dolldrop_.pull_to_path(uri, droppath);
}

//------------------------------------------------------------------------------
auto Bumpmaster::pull_hdr_body(const std::string& uri) -> std::stringstream {
  return dolldrop_.pull_hdr_body(uri);
}

//------------------------------------------------------------------------------
auto Bumpmaster::top_bump_thrds(Json const&        cat_json,
                                std::string const& cat_uri,
                                fs::path const&    board_jsons_path,
                                const int thrd_lim) -> std::vector<Thrd_data> {
  stamp("top_bump_thrds()");

  vector<Thrd_data> res;

  // store all thread's data (msg, subj, ID) from the catalog; in bump-order
  auto thrd_dats = bump_thrd_dats(cat_json);

  // assemble all thread's JSON page URIs; in bump-order
  auto const& [site_nm, board_nm] = parse_ib_uri(cat_uri);
  auto const thrd_uris            = bump_thrd_uris(cat_json, site_nm, board_nm);

  for (auto t_idx{0}; t_idx < thrd_lim; ++t_idx) {  // for all Page 1 threads,
                                                    //
    auto const& thrd_uri = thrd_uris[t_idx];        //  index to thread URI
                                                    //
    auto& thrd_dat = thrd_dats[t_idx];              //  index to thread data

    // assemble each thread's local JSON filesystem path
    auto const thrd_nm = to_string(thrd_dat.id) + ".json";
    auto const thrd_path{board_jsons_path / thrd_nm};

    cout << " page_uri:    " << thrd_uri << '\n';
    auto thrd_json = dolldrop_.pull_to_path(thrd_uri, thrd_path);

    thrd_dat.posts = thrd_posts(thrd_json);  // copy post's data into cont

    res.push_back(thrd_dat);
  }

  return res;
}

#endif  // __has_include(<curl_easy.h>)

}  // namespace rw::curator::bumpmaster

//    --=^=--

// "______"

// rw_curator_lib.v220507
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
