//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dolldrop.cpp
// -Definitions of the baseline RW Dolldrop sub-library

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "dolldrop_util.cpp"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
Dolldrop_b::Dolldrop_b(Hip const& tag, std::string const& handle)
    : RW_slave{{Type::Bumpmaster, Type::Curator}, handle, Type::Dolldrop},
      dollnet_{t(), "dolldrop_dollnet"} {
  stamp("Dolldrop_b()");

  bind_master(tag);

  preflight();
}

//------------------------------------------------------------------------------
auto Dolldrop_b::chkd_pull_g(Hip const&         tag,
                             curl_easy&         conn,
                             std::string const& uri,
                             fs::path const&    droppath) -> std::fstream {
  stamp(tag, "chkd_pull_g()");

  try {
    // pull the dolldrop's header
    auto hdr = pull_hdr_g(tag, conn, uri);

    // check if the remote source (claims to be) modified since last download
    // DESIGN: 'name' now unused, drop from called func ret
    //
    const auto& [have_new_drop, name] = chk_droptimes(uri, hdr);

    // check if the remote source and local dolldrop sizes match
    const auto size_match = chk_dropsizes(uri, hdr);

    if (have_new_drop) {  // yes, remote 'last-modified' header tag differs

      // -if this is a catalog's pull check, just a newer remote file is good
      // enough (since its size may stay identical, even w/ a legitimate change)
      // -otherwise, also require that the size of the remote differs from local
      //
      if (is_catalog_ || (! size_match)) {  // yes, do request the dolldrop pull

        // adjust the display cursor back up one line -re: prelim header request
        cout << "\u008D\r" << flush;

        return pull_body_g(tag, conn, uri, droppath);
      }
    }

    // otherwise, just return the already-saved (and current) local dolldrop
    return fstream{droppath};

  } catch (runtime_error const& e) {
    cerr << "\n  runtime error caught in chkd_pull_g()\n\n";
    return fstream{};

  } catch (...) {
    cerr << "\n  unknown error caught in chkd_pull_g()\n\n";
    return fstream{};
  }
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdr_g(Hip const&         tag,
                            curl_easy&         conn,
                            std::string const& uri) -> std::stringstream {
  stamp(tag, "pull_hdr_g()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    try {
      return dollnet_.dl_hdr(t(), conn, uri);

    } catch (runtime_error const& e) {
      cerr << "\n  runtime error caught in pull_hdr_g()\n\n";
      return stringstream{};

    } catch (...) {
      cerr << "\n  unknown error caught in pull_hdr_g()\n\n";
      return stringstream{};
    }

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdr_g()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_body_g(Hip const&         tag,
                             curl_easy&         conn,
                             std::string const& uri,
                             fs::path const&    droppath) -> std::fstream {
  stamp(tag, "pull_body_g()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    set_name(uri, droppath);

    ofstream anons_file{droppath};

    try {
      dollnet_.dl_body(t(), conn, uri, anons_file);

    } catch (runtime_error const& e) {
      cerr << "\n  runtime error caught in pull_body_g()\n\n";
      anons_file.close();
      return fstream{};

    } catch (...) {
      cerr << "\n  unknown error caught in pull_body_g()\n\n";
      anons_file.close();
      return fstream{};
    }

    anons_file.close();

    // TEMP: return dolldrop body directly for now
    return fstream{droppath};

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_body_g()");

  ///
  return fstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdr_body_g(Hip const&         tag,
                                 curl_easy&         conn,
                                 std::string const& uri) -> std::stringstream {
  stamp(tag, "pull_hdr_body_g()");

  // TEST:
  prt_tag(tag, "asked to pull a dolldrop w/ hdr");
  cout << "  -requested pull address :  " << uri << "\n\n";

  if (tag == master()) {  // function is meant only for slave's Master to use
    return try_hdr_body(t(), conn, uri);

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdr_body_g()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::try_hdr_body(Hip const&         tag,
                              curl_easy&         conn,
                              std::string const& uri) -> std::stringstream {  //

  auto const   dropname = bld_dropname();
  ofstream     anons_file{dropname};
  stringstream hdr;

  try {
    hdr = dollnet_.dl_hdr_body(tag, conn, uri, anons_file);

  } catch (runtime_error const& e) {
    anons_file.close();
    set_name(uri, dropname);
    return stringstream{};

  } catch (...) {
    cerr << "\n  unknown error caught in try_hdr_body()\n\n";
    anons_file.close();
    set_name(uri, dropname);
    return stringstream{};
  }

  anons_file.close();
  set_name(uri, dropname);

  // TEMP: return dolldrop header directly for now
  return hdr;
}

//------------------------------------------------------------------------------
auto Dolldrop_b::get_curl() -> curl_easy {
  return dollnet_.get_curl();
}

//------------------------------------------------------------------------------
auto Dolldrop_b::mk_dropdir() -> bool {
  fs::create_directory(dropdir);
  return exists_fs(dropdir);
}

//------------------------------------------------------------------------------
void Dolldrop_b::wrt_tailme() {  //

  // a raw string; the body of the text file
  auto const tailtext =
      R"(//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

Notes:

-You can use the tail command to watch DLs progressively fill your dolldrops
 E.G., (in a terminal from the project's base directory) :

    tail -f  ./dolldrops/anons_file_001

-You can even playback a stream-capp style dolldrop 'live' (while its capturing)
 E.G. :

    vlc  ./dolldrops/anons_file_001
       or
    mpv  ./dolldrops/anons_file_042.ogg

-Protip : If you want a file extension automatically appended to your dolldrops,
 put it on the second line of the file named :  ./dolldrops/tagfile
 E.G., for OGG dolldrops, a tagfile might be edited to begin with the contents :

    000
    .ogg

 which would (on repeated use of this demo) create a series of dolldrops named :

    ./dolldrops/anons_file_001.ogg
    ./dolldrops/anons_file_002.ogg
    ./dolldrops/anons_file_003.ogg

 &tc. (btw this can be changed anytime to get a different extension next round)

Cheers.

// rw_dollhouse_lib
// Copyright (C)
// License MIT (Expat) https://opensource.org/licenses/MIT
)";

  ofstream ofs{dropdir / "tail_me.txt"};
  ofs << tailtext;
  ofs.close();
}

//------------------------------------------------------------------------------
void Dolldrop_b::preflight() {  //

  if (! exists_fs(dropdir)) {  // check if ./dolldrops/ directory exists

    try {
      mk_dropdir();
      wrt_tailme();

    } catch (fs::filesystem_error const& e) {
      cout << "  fs::filesystem_error : '" << e.what()
           << "' caught in Mailstop::preflight()\n\n";

    } catch (std::bad_alloc const& e) {
      cout << "  std::bad_alloc : '" << e.what()
           << "' caught in Mailstop::preflight()\n\n";
    }
  }
}

//------------------------------------------------------------------------------
auto Dolldrop::get_ib_curl(std::string const& uri) -> curl_easy {
  auto const& [site_nm, board_nm] = parse_ib_uri(uri);

  // obtain the previous curl_easy obj for this site if found, else create new
  curl_easy conn;
  try {
    conn = site_conns_.at(site_nm);

  } catch (out_of_range const& e) {
    conn                 = get_curl();
    site_conns_[site_nm] = conn;
  }

  return conn;
}

//------------------------------------------------------------------------------
auto Dolldrop::pull_to_path(std::string const& uri, fs::path const& droppath)
    -> Json {
  auto conn = get_ib_curl(uri);

  chkd_pull_g(master(), conn, uri, droppath);  // ignore ret, just pull

  return rd_json_file(droppath);
}

//------------------------------------------------------------------------------
auto Dolldrop::pull_hdr_body(std::string const& uri) -> std::stringstream {
  auto conn = get_ib_curl(uri);

  return pull_hdr_body_g(master(), conn, uri);
}

//------------------------------------------------------------------------------
void Dolldrop::do_catalog(bool const flag) {
  stamp("do_catalog()");

  is_catalog_ = flag;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dolldrop_lib.v220507
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
